﻿using UnityEngine;
using UnityEngine.UI;

using System.Collections;
using System.Collections.Generic;
using System.IO;

public class timeline_controller : MonoBehaviour
{
	public static timeline_controller _singleton;

	public GameObject keyframe_prefab;
	public GameObject keyframe_index_prefab;
	public GameObject initial_keyframe_index;
	public GameObject canvas;
	public Camera cam;
	public GameObject keyframe_highlight;
	private RawImage keyframe_raw;
	public Texture alt_highlight;

	public GameObject user;
	public GameObject rabbit;
	public GameObject squirrel;
	//public GameObject chinchilla;

	public GameObject keyframe_container;

	private float frame_id = -0.5f;

	private List<GameObject> keyframes;

	private List<Vector3> rabbit_positions;
	private List<Vector3> rabbit_rotations;
	private List<Vector3> squirrel_positions;
	private List<Vector3> squirrel_rotations;
	//private List<Vector3> chinchilla_positions;
	//private List<Vector3> chinchilla_rotations;

	private List<Vector3> user_positions;
	private List<Vector3> user_rotations;

	private List<GameObject> keyframe_indices;

	public GameObject manip_obj;
	private manip manip_script;

	public GameObject camera_obj;
	private OVRScreenFade fader;

	private bool is_trigger_down = false;

	public RawImage sample_timeline;

	void Awake ()
	{
		if (_singleton == null) {
			DontDestroyOnLoad (gameObject);
			_singleton = this;
		} else if (_singleton != this) {
			Destroy (this);
		}

		manip_script = manip_obj.GetComponent<manip> ();
		fader = camera_obj.GetComponent<OVRScreenFade> ();
	}

	void Start ()
	{
		keyframes = new List<GameObject> ();

		rabbit_positions = new List<Vector3> ();
		rabbit_rotations = new List<Vector3> ();
		squirrel_positions = new List<Vector3> ();
		squirrel_rotations = new List<Vector3> ();
		//chinchilla_positions = new List<Vector3> ();
		//chinchilla_rotations = new List<Vector3> ();

		user_positions = new List<Vector3> ();
		user_rotations = new List<Vector3> ();

		keyframe_indices = new List<GameObject> ();

		keyframe_raw = (RawImage)keyframe_highlight.GetComponent<RawImage> ();
	}

	void Update ()
	{
		if (manip._singleton.inputMode == manip.inputModes.NORMAL) {
			ListenForXboxInput ();
		}		
	}

	void ListenForXboxInput ()
	{
		float triggers = -Input.GetAxis ("Triggers");
		float zoom = Input.GetAxis ("LDPadVertical");

		// If triggers are no longer being pressed allow user to jump keyframes again
		if (triggers == 0.0f)
			is_trigger_down = false;

		// Capture keyframe
		if (Input.GetKeyUp (KeyCode.JoystickButton0) || Input.GetKeyUp (KeyCode.M)) {
			StartCoroutine (CaptureKeyframe());
		}

		// Delete keyframe
		if (Input.GetKeyUp (KeyCode.JoystickButton1) || Input.GetKeyUp (KeyCode.I)) {
			if (OnKeyframe()) {
				DeleteKeyframe (keyframes[(int) frame_id], keyframe_indices[(int) frame_id], false);
				if (keyframes.Count > 0)
					UpdateObj ();
				manip_script.ResetGhosts();
			}
		}
		
		// Next keyframe/gap
		if (Input.GetKeyUp (KeyCode.JoystickButton5) || Input.GetKeyUp (KeyCode.K)) {
			if (frame_id < (float)keyframes.Count - 0.5f) { 
				ShiftLeft (keyframes.Count);
				frame_id += 0.5f;
				Debug.Log(frame_id);
				if (OnKeyframe()) {
					UpdateObj ();
					manip_script.ResetGhosts();
				}
				ChangeHighlight();
			}
		}
		
		// Previous keyframe/gap
		if (Input.GetKeyUp (KeyCode.JoystickButton4) || Input.GetKeyUp (KeyCode.J)) {
			if (frame_id >= 0.0f) {
				ShiftRight (0);
				frame_id -= 0.5f;
				Debug.Log(frame_id);
				if (OnKeyframe()) {
					UpdateObj ();
					manip_script.ResetGhosts();
				}
				ChangeHighlight();
			}
		}

		// Reset user position
		if (Input.GetKeyUp (KeyCode.JoystickButton2)) {
			if (OnKeyframe())
				StartCoroutine (ResetUser());
		}

		// Next keyframe
		if (triggers > 0.0f && !is_trigger_down) {
			if (OnKeyframe() && frame_id < (float)keyframes.Count - 1.0f) {
				is_trigger_down = true;
				ShiftLeftDouble (keyframes.Count);
				frame_id += 1.0f;
				Debug.Log(frame_id);
				UpdateObj ();
				manip_script.ResetGhosts();
			}
		}

		// Previous keyframe
		if (triggers < 0.0f && !is_trigger_down) {
			if (OnKeyframe() && frame_id > 0.0f) {
				is_trigger_down = true;
				ShiftRightDouble (0);
				frame_id -= 1.0f;
				Debug.Log(frame_id);
				UpdateObj ();
				manip_script.ResetGhosts();
			}
		}

		// Display sample timeline
		if (Input.GetKey (KeyCode.JoystickButton8)) {
			sample_timeline.enabled = true;
		} else {
			sample_timeline.enabled = false;
		}

		// Change zoom
		if (zoom < 0.0f) {
			cam.fieldOfView -= 1.0f;
		}
		if (zoom > 0.0f) {
			cam.fieldOfView += 1.0f;
		}
	}

	IEnumerator CaptureKeyframe ()
	{
		if (look_input._singleton.target != null) {
			look_input._singleton.target.GetComponent<Renderer> ().material.color = Color.white;
		}

		// Display on timeline
		GameObject keyframe = (GameObject)Instantiate (keyframe_prefab, new Vector3 (0.0f, 0.35f, 0), Quaternion.identity);

		if (OnKeyframe() && (int)frame_id < keyframes.Count) {
			Debug.Log("Overwrite");

			if (keyframes.Count > 0)
				DeleteKeyframe (keyframes[(int) frame_id], null, true);
			keyframes.Insert ((int)frame_id, keyframe);
			manip_script.ResetGhosts();

		} else if (!OnKeyframe()) {
			Debug.Log("Insert");

			frame_id = Mathf.Ceil(frame_id);

			// Create keyframe index text
			GameObject keyframe_index_text = (GameObject)Instantiate (keyframe_index_prefab);
			int keyframe_index = (int)frame_id + 1;
			keyframe_index_text.GetComponent<Text>().text = keyframe_index.ToString();

			if ((int) frame_id >= 0 && (int) frame_id < keyframes.Count) {
				keyframes.Insert ((int)frame_id, keyframe);
				keyframe_indices.Insert ((int)frame_id, keyframe_index_text);
			} else {
				keyframes.Add (keyframe);
				keyframe_indices.Add (keyframe_index_text);
			}

			UpdateIndicesText();

			//Enable/Disable initial keyframe number text
			if (keyframes.Count >= 1)
				initial_keyframe_index.GetComponent<Text>().enabled = false;
			else
				initial_keyframe_index.GetComponent<Text>().enabled = true;

			keyframe_index_text.transform.SetParent (canvas.transform, false);
			CreateGap((int)frame_id);
		}
		keyframe.transform.SetParent (keyframe_container.transform, false);
		Debug.Log ("at: " + frame_id);

		// Record object and user positions
		Vector3 rabbit_pos = new Vector3 (rabbit.transform.position.x, rabbit.transform.position.y, rabbit.transform.position.z);
		rabbit_positions.Insert ((int)frame_id, rabbit_pos);
		Vector3 rabbit_rot = new Vector3 (rabbit.transform.eulerAngles.x, rabbit.transform.eulerAngles.y, rabbit.transform.eulerAngles.z);
		rabbit_rotations.Insert ((int)frame_id, rabbit_rot);

		Vector3 squirrel_pos = new Vector3 (squirrel.transform.position.x, squirrel.transform.position.y, squirrel.transform.position.z);
		squirrel_positions.Insert ((int)frame_id, squirrel_pos);
		Vector3 squirrel_rot = new Vector3 (squirrel.transform.eulerAngles.x, squirrel.transform.eulerAngles.y, squirrel.transform.eulerAngles.z);
		squirrel_rotations.Insert ((int)frame_id, squirrel_rot);

		//Vector3 chinchilla_pos = new Vector3 (chinchilla.transform.position.x, chinchilla.transform.position.y, chinchilla.transform.position.z);
		//chinchilla_positions.Insert ((int)frame_id, chinchilla_pos);
		//Vector3 chinchilla_rot = new Vector3 (chinchilla.transform.eulerAngles.x, chinchilla.transform.eulerAngles.y, chinchilla.transform.eulerAngles.z);
		//chinchilla_rotations.Insert ((int)frame_id, chinchilla_rot);

		Vector3 user_pos = new Vector3 (user.transform.position.x, user.transform.position.y, user.transform.position.z);
		user_positions.Insert ((int)frame_id, user_pos);
		Vector3 user_rot = new Vector3 (user.transform.eulerAngles.x, user.transform.eulerAngles.y, user.transform.eulerAngles.z);
		user_rotations.Insert ((int)frame_id, user_rot);

		// Turn off UI for screenshot
		cam.cullingMask ^= 1 << LayerMask.NameToLayer ("UI");
		yield return new WaitForSeconds (0.01f);
		Application.CaptureScreenshot ("Assets/Keyframes/keyframe_" + frame_id + ".png");
		// Turn back on
		yield return new WaitForSeconds (0.01f);
		cam.cullingMask ^= 1 << LayerMask.NameToLayer ("UI");
		
		// Update texture
		RawImage keyframe_img = keyframe.GetComponent<RawImage> ();
		string filepath = Application.dataPath + "/Keyframes/keyframe_" + frame_id + ".png";
		Texture2D tex = new Texture2D (1, 1);
		tex.LoadImage (System.IO.File.ReadAllBytes (filepath));
		keyframe_img.texture = tex;

		if (look_input._singleton.target != null) {
			look_input._singleton.target.GetComponent<Renderer> ().material.color = Color.red;
		}

		if (keyframes.Count > 0)
			UpdateObj ();

		ChangeHighlight();
		
	}

	void DeleteKeyframe (GameObject keyframe, GameObject keyframe_index, bool overwriting)
	{
		if (keyframe != null) {
			keyframes.RemoveAt ((int) frame_id);
			
			rabbit_positions.RemoveAt ((int) frame_id);
			rabbit_rotations.RemoveAt ((int) frame_id);
			squirrel_positions.RemoveAt ((int) frame_id);
			squirrel_rotations.RemoveAt ((int) frame_id);
			//chinchilla_positions.RemoveAt ((int) frame_id);
			//chinchilla_rotations.RemoveAt ((int) frame_id);
			
			user_positions.RemoveAt ((int) frame_id);
			user_rotations.RemoveAt ((int) frame_id);
			
			Destroy (keyframe);

			if (keyframe_index != null) {
				keyframe_indices.RemoveAt ((int) frame_id);
				Destroy (keyframe_index);
			}

			//Readjust timeline
			if (!overwriting) {
				if ((int) frame_id >= keyframes.Count) {
					frame_id = keyframes.Count - 0.5f;
					ShiftRight (0);
				} else {
					ShiftLeftAlt ((int) frame_id);
					ShiftLeftAlt ((int) frame_id);
				}
			}

			UpdateIndicesText();
			ChangeHighlight();

			if (keyframes.Count == 0) {
				initial_keyframe_index.GetComponent<Text>().enabled = true;
				keyframe_raw.texture = null;
			}
		}
	}

	void ShiftLeft (int bound)
	{
		if (bound >= 0 && bound <= keyframes.Count) {
			for (int i = 0; i < bound; i++) {
				//Keyframe
				Vector3 pos = keyframes [i].transform.localPosition;
				Vector3 newpos = new Vector3 (pos.x - (0.325f / 2.0f), 0.35f, 0.0f);
				keyframes [i].transform.localPosition = newpos;

				//Keyframe index text
				Vector3 text_pos = keyframe_indices [i].transform.localPosition;
				Vector3 text_newpos = new Vector3 (text_pos.x - (0.325f / 2.0f), 0.4f, 0.0f);
				keyframe_indices [i].transform.localPosition = text_newpos;
			}
		}
	}

	void ShiftLeftDouble (int bound)
	{
		if (bound >= 0 && bound <= keyframes.Count) {
			for (int i = 0; i < bound; i++) {
				//Keyframe
				Vector3 pos = keyframes [i].transform.localPosition;
				Vector3 newpos = new Vector3 (pos.x - 0.325f, 0.35f, 0.0f);
				keyframes [i].transform.localPosition = newpos;
				
				//Keyframe index text
				Vector3 text_pos = keyframe_indices [i].transform.localPosition;
				Vector3 text_newpos = new Vector3 (text_pos.x - 0.325f, 0.4f, 0.0f);
				keyframe_indices [i].transform.localPosition = text_newpos;
			}
		}
	}

	void ShiftRight (int bound)
	{
		if (bound >= 0 && bound <= keyframes.Count) {
			for (int i = bound; i < keyframes.Count; i++) {
				Vector3 pos = keyframes [i].transform.localPosition;
				Vector3 newpos = new Vector3 (pos.x + (0.325f / 2.0f), 0.35f, 0.0f);
				keyframes [i].transform.localPosition = newpos;

				Vector3 text_pos = keyframe_indices [i].transform.localPosition;
				Vector3 text_newpos = new Vector3 (text_pos.x + (0.325f / 2.0f), 0.4f, 0.0f);
				keyframe_indices [i].transform.localPosition = text_newpos;
			}
		}
	}

	void ShiftRightDouble (int bound)
	{
		if (bound >= 0 && bound <= keyframes.Count) {
			for (int i = bound; i < keyframes.Count; i++) {
				Vector3 pos = keyframes [i].transform.localPosition;
				Vector3 newpos = new Vector3 (pos.x + 0.325f, 0.35f, 0.0f);
				keyframes [i].transform.localPosition = newpos;
				
				Vector3 text_pos = keyframe_indices [i].transform.localPosition;
				Vector3 text_newpos = new Vector3 (text_pos.x + 0.325f, 0.4f, 0.0f);
				keyframe_indices [i].transform.localPosition = text_newpos;
			}
		}
	}

	void ShiftLeftAlt (int bound)
	{
		if (bound >= 0 && bound <= keyframes.Count) {
			for (int i = bound; i < keyframes.Count; i++) {
				Vector3 pos = keyframes [i].transform.localPosition;
				Vector3 newpos = new Vector3 (pos.x - (0.325f / 2.0f), 0.35f, 0.0f);
				keyframes [i].transform.localPosition = newpos;

				Vector3 text_pos = keyframe_indices [i].transform.localPosition;
				Vector3 text_newpos = new Vector3 (text_pos.x - (0.325f / 2.0f), 0.4f, 0.0f);
				keyframe_indices [i].transform.localPosition = text_newpos;
			}
		}
	}

	void CreateGap (int frame_id)
	{
		if (keyframes.Count > 1) {
			ShiftLeft (frame_id);
			ShiftRight (frame_id + 1);
		}
	}

	void UpdateObj ()
	{
		rabbit.transform.position = rabbit_positions [(int)frame_id];
		rabbit.transform.eulerAngles = rabbit_rotations [(int)frame_id];

		squirrel.transform.position = squirrel_positions [(int)frame_id];
		squirrel.transform.eulerAngles = squirrel_rotations [(int)frame_id];

		//chinchilla.transform.position = chinchilla_positions [(int)frame_id];
		//chinchilla.transform.eulerAngles = chinchilla_rotations [(int)frame_id];
	}

	IEnumerator ResetUser ()
	{
		fader.StartFadeOut ();

		yield return new WaitForSeconds (2.0f);
		
		fader.StartFadeIn ();

		user.transform.position = user_positions [(int)frame_id];
		user.transform.eulerAngles = user_rotations [(int)frame_id];		
	}

	public bool OnKeyframe ()
	{
		string str_frame = frame_id.ToString ();
		return !(str_frame.Contains ("."));
	}

	void ChangeHighlight ()
	{
		if (!OnKeyframe ())
			keyframe_raw.texture = (Texture)alt_highlight;
		else
			keyframe_raw.texture = null;
	}

	void UpdateIndicesText ()
	{
		for (int i = 0; i < keyframe_indices.Count; i++) {
			keyframe_indices[i].GetComponent<Text>().text = (i + 1).ToString();
		}
	}
}
